
xVersion" 2011.11.5.1"

module timers

  #Document timers
  //
  //  Generic timer support for 16 bit timers, for PIC24
  //
  //    Version 2, 2008-02-29
  //
  //
  //  When including this file in make.4th and using this module by means of
  //  the uses directive (uses Timers endUses)) Timers can be defined in the
  //  following way :
  //
  //    [[timer]] TimerName   // ( -- )
  //
  //
  //  To integrate the timer into the system somewhere in a timer interrupt
  //  (likely ...) a call to the timer's interrupt handler must be made. Like :
  //
  //    call    <@" TimerName.tick">
  //
  //  Or for a 4th interrupt just :
  //
  //    TimerName.tick
  //
  //
  //  The interrupt call determines the timer's run-rate, for a 1 ms timer the
  //  tick method should be called every ms.
  //
  //
  //  Some X4th words will be auto defined (but be only compiled if actually
  //  used) for each timer :
  //
  //    TimerName
  //
  //      An autodefined 16 bit variable used to count the time down.
  //
  //
  //    TimerNamecontrol
  //
  //      An autodefined 16 bit variable used to hold two control bits for the
  //      the timer.
  //
  //        bit0 - [[TimerRuns]]  - the timer runs when this bit is set.
  //        bit1 - [[TimerFired]] - when set indicates that the timer fired
  //
  //
  //    TimerName.start ( Intervals -- )
  //
  //      Intervals is a word (2 bytes)
  //
  //      Use as in 250 TimerName.start to start the timer with a run
  //      time o250 interrupt calls (ticks).
  //
  //
  //    Timername.stop ( -- )
  //
  //      Used to stop the timer and set it un-fired
  //
  //
  //    TimerName.fired ( -- f )
  //
  //      Used to check if the timer fired as in :
  //
  //      TimerName.fired
  //      if
  //        some_actions...
  //      endIf
  //
  //
  //    TimerName.runs ( -- f )
  //
  //      Can be used to check if the timer is running, as in :
  //
  //      TimerName.runs
  //      if
  //        some_actions...
  //      endIf
  //
  //
  //    TimerName.setfired ( -- )
  //
  //      Can be used to flag the timer as being fired.
  //
  //
  //    TimerName.tick ( -- )
  //
  //      The interrupt service routine for the timer. Meant to be called
  //      from a timed interrupt. Each call advances the timer by one time step
  //      setting it to be fired when it's time is over.
  //
  //      Meant to be called from a timer interrupt, but it can be used from
  //      non-interrupt code as well. It advances the timer one tick per call.
  //
  #endDoc



public


  // //////////////////////////////////////////////////////////////////////////
  // Bit definitions for <timer>Control word

  0 constant timerRuns                        #// [[Timers]] bit 0
  1 constant timerFired                       #// [[Timers]] bit 1


{ // start meta compiler


  // //////////////////////////////////////////////////////////////////////////
  // Primitve stuff, the assembly code dealing with the timers

  macro tmr.start ( aName ) ( wTime -- ) // [[Timers]] Start 16 bit timer
                                                ; // ( wTime -- )
    mov       #<@" $aName$">Control, w1         ; // Point to timer control word
    clr       [w1]                              ; // Clear run and fired bits
    PopReg    <@" $aName$">                     ; // Pop interval into timer register
    bset      [w1], #<@" timerRuns">            ; // Start the timer
  endMacro


  // //////////////////////////////////////////////////////////////////////////

  macro tmr.stop ( aName ) ( -- ) // [[Timers]] Stop timer
    mov       #<@" $aName$">Control, w1         ; // Point to timer control word
    clr       [w1]                              ; // Clear run and fired bits
  endMacro


  // //////////////////////////////////////////////////////////////////////////

  macro tmr.fired ( aName ) ( -- f ) // [[Timers]] Check if timer fired
    mov       #<@" $aName$">Control, w1         ; // Point to timer control word
    PushNothing                                 ; // Make room on stack
                                                ; //   for f.
    mov       #( 1 << <@" timerFired">), w0     ; // Set Zero flag if timer did
                                                ; //   not fire.
    and       w0, [w1], w0
    bra       z, <@" $aName$">_skip             ; // B/ Z, timer did not
                                                ; //       fire
    clr       [w1]                              ; // Clear run bit and fired bit
                                                ; //   so we see it once only.
<@" $aName$">_skip:

    and       #( 1 << <@" timerFired">), w0     ; // Restore Z bit
  endMacro


  // //////////////////////////////////////////////////////////////////////////

  macro tmr.runs ( aName ) ( -- f ) // [[Timers]] Check if timer runs
    mov       #<@" $aName$">Control, w1         ; // Point to timer control word
    PushNothing                                 ; // Make room on stack for f
    mov       #( 1 << <@" timerRuns">)          ; // Set Zero flag if timer does
                                                ; // not run.
    and       [w1], w0
  endMacro


  // //////////////////////////////////////////////////////////////////////////

  macro tmr.setFired ( aName ) ( -- ) // [[Timers]] Set timer fired
    mov       #<@" $aName$">Control, w1         ; // Point to timer control word
    bclr      [w1], #<@" timerRuns">            ; // Stop the timer
    bset      [w1], #<@" timerFired">           ; // Set Fired bit
  endMacro


  // //////////////////////////////////////////////////////////////////////////

  macro tmr.tick ( aName ) ( -- ) // [[Timers]] Interrupt handler for 16 bit timer

    ; // //////////////////////////////////////////////////////////////////////
    ; // timer : $aName$ - 16 bits
    ; // //////////////////////////////////////////////////////////////////////

    mov       #<@" $aName$">Control, w1         ; // Get address of timer control word into w1
    mov       #( 1 << <@" timerRuns">), w0      ; // Set Zero flag if
                                                ; //   timer does not
                                                ; //   run.
    and       w0, [w1], w0
    bra       z, intskp_<@" $aName$">           ; // B/ Timer stopped,
                                                ; //   ignore it.

    ; // Timer rums

    mov       #( 1 << <@" timerFired">), w0     ; // Set Zero flag if
                                                ; //   timer did not
                                                ; //   fire.
    and       w0, [w1], w0
    bra       z, <@" $aName$">_int_skip         ; // B/ Z, timer did not
                                                ; //       fire.

    clr       [w1]                              ; // Clear run bit and
                                                ; //    fired bit so we
                                                ; //    see it once only.
<@" $aName$">_int_skip:

    and       #( 1 << <@" timerFired">), w0     ; // Restore Z bit
    bra       nz, intskp_<@" $aName$">          ; // Timer already fired,
                                                ; //   ignore it

    ; // Timer runs and did not fire yet

    dec       <@" $aName$">                     ; // Decrement timer
    bra       nz, intskp_<@" $aName$">          ; // B/ timer <> 0

    ; // Timer runs and just fired

    bclr      [w1], #<@" timerRuns">            ; // Stop the timer
    bset      [w1], #<@" timerFired">           ; // Set Fired bit

    ; // Fall through to intskp$aName$ : check next timer

intskp_<@" $aName$">:

  endMacro


  // //////////////////////////////////////////////////////////////////////////
  // Macro's to define the x4th interface - parametrized
  // This defines a set of colon words that can be used either from x4th
  // or from assembly code. These are defined as meta words.

    macro timer ( aName ) ( -- )
    <!
      variable  $aName$               #//  [[Timers]] 16 bit counter
      variable  $aName$Control        #//  [[Timers]] control word

      : $aName$.Start ( bTime -- )

        #//  [[Timers]] Start the timer with a 16 bit tick value

        tmr.start $aName$
      ;


      : $aName$.stop ( -- )

        #//  [[Timers]] Stop the timer, make it not running and not fired

        tmr.stop $aName$
      ;


      : $aName$.fired ( -- f )

        #//  [[Timers]] Returns true when the timer fired, false otherwise

        tmr.fired $aName$
      ;


      : $aName$.runs ( -- f )

        #//  [[Timers]] Returns true when the timer is running, false otherwise

        tmr.runs $aName$
      ;


      : $aName$.setFired ( -- )

        #//  [[Timers]] Set the timer to be fired

        tmr.setFired $aName$
      ;


      : $aName$.tick ( -- )

        #//  [[Timers]] Interrupt handler for the timer

        tmr.tick $aName$
      ;
        resources
          $aName$
          $aName$Control
        endResources
    >
    endMacro
    #//  Used to define 16 bit [[Timers]]


} // end meta compiler

// end
// ////////////////////////////////////////////////////////////////////////////


endModule

